@@ -9,9 +9,16 @@
 
 #include "mmap.h"
 
+#ifndef WIN32
 #ifdef HAVE_MMAP
 #  include <sys/mman.h>
 #endif
+#else
+#include <ctype.h>
+#include <stdio.h>
+#include <windows.h>
+#undef HAVE_MADVISE
+#endif
 
 typedef struct {
   char * data;
@@ -116,10 +123,36 @@ SEXP mmap_mkFlags (SEXP _flags) {
       warning("unknown constant: skipped");
     }
   }
+#ifdef WIN32
+  flags_bit = PAGE_READWRITE;
+  if(flags_bit & PROT_READ & PROT_WRITE & MAP_SHARED)
+    flags_bit = PAGE_READWRITE;
+  else if(flags_bit & MAP_PRIVATE)
+    flags_bit = PAGE_WRITECOPY;
+  else if(flags_bit & PROT_READ)
+    flags_bit = PAGE_READONLY;
+#endif
   return ScalarInteger(flags_bit);
 } /*}}}*/
 
 /* mmap_munmap {{{ */
+#ifdef WIN32
+SEXP mmap_munmap (SEXP mmap_obj) {
+  int ret;
+  char *data = MMAP_DATA(mmap_obj);
+  HANDLE fd = (HANDLE)MMAP_FD(mmap_obj);
+  HANDLE mh = (HANDLE)MMAP_HANDLE(mmap_obj);
+
+  if(data == NULL)
+    error("invalid mmap pointer");
+
+  ret = UnmapViewOfFile(data);
+  CloseHandle(mh);
+  CloseHandle(fd);
+  R_ClearExternalPtr(VECTOR_ELT(mmap_obj,0));
+  return(ScalarInteger(ret));
+}
+#else
 SEXP mmap_munmap (SEXP mmap_obj) {
   char *data = MMAP_DATA(mmap_obj);
   int fd = MMAP_FD(mmap_obj);
@@ -132,8 +165,47 @@ SEXP mmap_munmap (SEXP mmap_obj) {
   R_ClearExternalPtr(VECTOR_ELT(mmap_obj,0));
   return(ScalarInteger(ret)); 
 } /*}}}*/
+#endif
 
 /* mmap_mmap AND mmap_finalizer {{{ */
+#ifdef WIN32
+SEXP mmap_mmap (SEXP _type, SEXP _fildesc, SEXP _prot,
+                SEXP _flags, SEXP _len, SEXP _off) {
+  char *data;
+  struct stat st;
+  SYSTEM_INFO sSysInfo;
+  GetSystemInfo(&sSysInfo);
+
+  stat(CHAR(STRING_ELT(_fildesc,0)), &st);
+
+  HANDLE hFile=CreateFile(CHAR(STRING_ELT(_fildesc,0)),
+                  GENERIC_READ|GENERIC_WRITE,
+                  FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,NULL);
+
+  HANDLE hMap=CreateFileMapping(hFile,NULL,PAGE_READWRITE,0,0,NULL);
+  DWORD dwFileSize=GetFileSize(hFile,NULL);
+  data = (char *)MapViewOfFile(hMap,FILE_MAP_WRITE,0,0,dwFileSize);
+
+
+  SEXP mmap_obj;
+  PROTECT(mmap_obj = allocVector(VECSXP,6));
+  /* data pointer    */
+  SET_VECTOR_ELT(mmap_obj, 0, R_MakeExternalPtr(data,R_NilValue,R_NilValue));
+  /* size in bytes   */
+  SET_VECTOR_ELT(mmap_obj, 1, _len);
+  /* file descriptor */
+  SET_VECTOR_ELT(mmap_obj, 2, ScalarInteger((int)hFile));
+  /* storage mode    */
+  SET_VECTOR_ELT(mmap_obj, 3, _type);
+  /* page size       */
+  SET_VECTOR_ELT(mmap_obj, 4, ScalarReal((double)sSysInfo.dwPageSize));
+  /* Map handle */
+  SET_VECTOR_ELT(mmap_obj, 5, ScalarInteger((int)hMap));
+
+  UNPROTECT(1);
+  return(mmap_obj);
+}
+#else
 SEXP mmap_mmap (SEXP _type, SEXP _fildesc, SEXP _prot,
                 SEXP _flags, SEXP _len, SEXP _off) {
   int fd;
@@ -145,7 +217,7 @@ SEXP mmap_mmap (SEXP _type, SEXP _fildesc, SEXP _prot,
   if(fd < 0)
     error("unable to open file");
   data = mmap((caddr_t)0, 
-              INTEGER(_len)[0], 
+              (long)REAL(_len)[0], 
               INTEGER(_prot)[0], 
               INTEGER(_flags)[0], 
               fd, 
@@ -182,11 +254,20 @@ SEXP mmap_mmap (SEXP _type, SEXP _fildesc, SEXP _prot,
   UNPROTECT(1);
   return(mmap_obj);
 } /*}}}*/
+#endif
 
 /* mmap_pagesize {{{ */
+#ifdef WIN32
+SEXP mmap_pagesize () {
+  SYSTEM_INFO sSysInfo;
+  GetSystemInfo(&sSysInfo);
+  return ScalarInteger((int)sSysInfo.dwPageSize);
+} /*}}}*/
+#else
 SEXP mmap_pagesize () {
   return ScalarInteger((int)sysconf(_SC_PAGE_SIZE));
 } /*}}}*/
+#endif
 
 /* mmap_is_mmapped {{{ */
 SEXP mmap_is_mmapped (SEXP mmap_obj) {
@@ -197,6 +278,14 @@ SEXP mmap_is_mmapped (SEXP mmap_obj) {
   return(ScalarLogical(1));
 } /*}}}*/
 
+#ifdef WIN32
+SEXP mmap_msync (SEXP mmap_obj, SEXP _flags) {
+  char *data;
+  data = MMAP_DATA(mmap_obj);
+  FlushViewOfFile((void *)data, (size_t)MMAP_SIZE(mmap_obj));
+  return 0;
+}/*}}}*/
+#else
 /* {{{ mmap_msync */
 SEXP mmap_msync (SEXP mmap_obj, SEXP _flags) {
   char *data;
@@ -204,6 +293,7 @@ SEXP mmap_msync (SEXP mmap_obj, SEXP _flags) {
   int ret = msync(data, MMAP_SIZE(mmap_obj), INTEGER(_flags)[0]);
   return ScalarInteger(ret);
 }/*}}}*/
+#endif
 
 /* {{{ mmap_madvise */
 SEXP mmap_madvise (SEXP mmap_obj, SEXP _len, SEXP _flags) {
@@ -247,7 +337,7 @@ Rprintf("offset: %i\n",(ival/pagesize)*pagesize);
 /* {{{ mmap_extract */
 SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
 /*SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {*/
-  int v, fi, i, ii, ival;
+  long v, fi, i, ii, ival;
   int P=0;
   unsigned char *data; /* unsigned int and values */
 
@@ -288,7 +378,7 @@ SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
   P++;
 
   int *index_p = INTEGER(index);
-  int upper_bound;
+  long upper_bound;
 
   /* need R typed storage for structures... 
      ideally we needn't alloc for types
@@ -305,21 +395,29 @@ SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
   switch(mode) {
   case INTSXP: /* {{{ */
     int_dat = INTEGER(dat);
-    upper_bound = (int)(MMAP_SIZE(mmap_obj)-Cbytes);
+    upper_bound = (long)(MMAP_SIZE(mmap_obj)-Cbytes);
     switch(Cbytes) {
       case 1: /* 1 byte (signed) char */
         if(isSigned) {
           for(i=0;  i < LEN; i++) {
             ival = (index_p[i]-1);
-            if( ival > upper_bound || ival < 0 )
+            if( ival > upper_bound || ival < 0 ) {
+              if( ival == 0 ) {
+                continue;
+              }
               error("'i=%i' out of bounds", index_p[i]);
+            }
             int_dat[i] = (int)(char)(data[(index_p[i]-1)]);
           }
         } else { /* unsigned */
           for(i=0;  i < LEN; i++) {
             ival = (index_p[i]-1);
-            if( ival > upper_bound || ival < 0 )
+            if( ival > upper_bound || ival < 0 ) {
+              if( ival == 0 ) {
+                continue;
+              }
               error("'i=%i' out of bounds", index_p[i]);
+             }
             int_dat[i] = (int)(unsigned char)(data[(index_p[i]-1)]);
           }
         }
@@ -328,8 +426,12 @@ SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
         if(isSigned) {
         for(i=0;  i < LEN; i++) {
           ival = (index_p[i]-1)*sizeof(short);
-          if( ival > upper_bound || ival < 0 )
+          if( ival > upper_bound || ival < 0 ) {
+            if( ival == 0 ) {
+              continue;
+            }
             error("'i=%i' out of bounds", index_p[i]);
+          }
           memcpy(short_buf, 
                  &(data[(index_p[i]-1)*sizeof(short)]),
                  sizeof(char)*sizeof(short));
@@ -338,8 +440,12 @@ SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
         } else {
         for(i=0;  i < LEN; i++) {
           ival = (index_p[i]-1)*sizeof(short);
-          if( ival > upper_bound || ival < 0 )
+          if( ival > upper_bound || ival < 0 ) {
+            if( ival == 0 ) {
+              continue;
+            }
             error("'i=%i' out of bounds", index_p[i]);
+          }
           memcpy(short_buf, 
                  &(data[(index_p[i]-1)*sizeof(short)]),
                  sizeof(char)*sizeof(short));
@@ -351,8 +457,12 @@ SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
         if(isSigned) {
         for(i=0;  i < LEN; i++) {
           ival =  (index_p[i]-1)*3;
-          if( ival > upper_bound || ival < 0 )
+          if( ival > upper_bound || ival < 0 ) {
+            if( ival == 0 ) {
+              continue;
+            }
             error("'i=%i' out of bounds", index_p[i]);
+          }
           memcpy(int24_buf, 
                  &(data[(index_p[i]-1)*3]), /* copy first 3 bytes */
                  3);
@@ -368,8 +478,12 @@ SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
         for(i=0;  i < LEN; i++) {
           ival =  (index_p[i]-1)*3;
           /* reset int_but to 0 0 0 0 */
-          if( ival > upper_bound || ival < 0 )
+          if( ival > upper_bound || ival < 0 ) {
+            if( ival == 0 ) {
+              continue;
+            }
             error("'i=%i' out of bounds", index_p[i]);
+          }
           memcpy(int24_buf, 
                  &(data[(index_p[i]-1)*3]), /* copy first 3 bytes */
                  3);
@@ -380,8 +494,12 @@ SEXP mmap_extract (SEXP index, SEXP field, SEXP mmap_obj) {
       case 4: /* 4 byte int */
         for(i=0;  i < LEN; i++) {
           ival =  (index_p[i]-1)*sizeof(int);
-          if( ival > upper_bound || ival < 0 )
+          if( ival > upper_bound || ival < 0 ) {
+            if( ival == 0 ) {
+              continue;
+            }
             error("'i=%i' out of bounds", index_p[i]);
+          }
           memcpy(int_buf, 
                  &(data[(index_p[i]-1)*sizeof(int)]),
                  sizeof(char)*sizeof(int));
